/*----- PROTECTED REGION ID(MCCE2Electrometers.cpp) ENABLED START -----*/
static const char *RcsId = "$Id:  $";
//=============================================================================
//
// file :        MCCE2Electrometers.cpp
//
// description : C++ source for the MCCE2Electrometers class and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               MCCE2Electrometers are implemented in this file.
//
// project :     Novelec MCCE-2 Electrometers
//
// This file is part of Tango device class.
//
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
//
// $Author:  $
//
// $Revision:  $
// $Date:  $
//
// $HeadURL:  $
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include <tango.h>
#include <PogoHelper.h>
#include <yat/utils/XString.h>
#include <yat4tango/InnerAppender.h>
#include <yat4tango/YatLogAdapter.h>
#include <MCCE2Electrometers.h>
#include <MCCE2ElectrometersClass.h>
#include "N_PhotoVoltaique.h"
#include "N_PhotoConducteur.h"
const size_t MAX_STRING_LENGTH = 64;
const unsigned short ELETRO_CHANNEL_ONE = 1;
const unsigned short ELETRO_CHANNEL_TWO = 2;

/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers.cpp

/**
 *  MCCE2Electrometers class description:
 *    This class allows you to control all MCCE-2 Novelec electrometers
 *    through a Serial bus.
 *    <br> Supported types :
 *    <br> Novelec Electrometers : MCCE2
 */

//================================================================
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name              |  Method name
//================================================================
//  State                     |  dev_state
//  Status                    |  dev_status
//  RangeUP                   |  range_up
//  RangeDOWN                 |  range_down
//  SetZeroVFFunction         |  set_zero_vffunction
//  SetOffsetZeroV1Function   |  set_offset_zero_v1_function
//  SetLeakageZeroV2Function  |  set_leakage_zero_v2_function
//  SetTestFunction           |  set_test_function
//  SetMeasureFunction        |  set_measure_function
//  MCCE2_ON                  |  mcce2__on
//  MCCE2_OFF                 |  mcce2__off
//  Local                     |  local
//  Remote                    |  remote
//  Reset                     |  reset
//  ClearRegisters            |  clear_registers
//  GetRange                  |  get_range
//  SetRange                  |  set_range
//  GetMode                   |  get_mode
//  GetPolarity               |  get_polarity
//  SetPolarity               |  set_polarity
//  GetHzFilter               |  get_hz_filter
//  SetHzFilter               |  set_hz_filter
//  GetGain                   |  get_gain
//  SetGain                   |  set_gain
//  GetElectrometerTypeStr    |  get_electrometer_type_str
//================================================================

//================================================================
//  Attributes managed are:
//================================================================
//  electrometerChannel  |  Tango::DevUShort	Scalar
//  range1               |  Tango::DevString	Scalar
//  range2               |  Tango::DevString	Scalar
//================================================================

namespace MCCE2Electrometers_ns
{
/*----- PROTECTED REGION ID(MCCE2Electrometers::namespace_starting) ENABLED START -----*/

//	static initializations

/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::namespace_starting

//--------------------------------------------------------
/**
 *	Method      : MCCE2Electrometers::MCCE2Electrometers()
 *	Description : Constructors for a Tango device
 *                implementing the classMCCE2Electrometers
 */
//--------------------------------------------------------
MCCE2Electrometers::MCCE2Electrometers(Tango::DeviceClass *cl, string &s)
 : Tango::Device_4Impl(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(MCCE2Electrometers::constructor_1) ENABLED START -----*/
	init_device();

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::constructor_1
}
//--------------------------------------------------------
MCCE2Electrometers::MCCE2Electrometers(Tango::DeviceClass *cl, const char *s)
 : Tango::Device_4Impl(cl, s)
{
	/*----- PROTECTED REGION ID(MCCE2Electrometers::constructor_2) ENABLED START -----*/
	init_device();

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::constructor_2
}
//--------------------------------------------------------
MCCE2Electrometers::MCCE2Electrometers(Tango::DeviceClass *cl, const char *s, const char *d)
 : Tango::Device_4Impl(cl, s, d)
{
	/*----- PROTECTED REGION ID(MCCE2Electrometers::constructor_3) ENABLED START -----*/
	init_device();

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::constructor_3
}

//--------------------------------------------------------
/**
 *	Method      : MCCE2Electrometers::delete_device()
 *	Description : will be called at device destruction or at init command
 */
//--------------------------------------------------------
void MCCE2Electrometers::delete_device()
{
	DEBUG_STREAM << "MCCE2Electrometers::delete_device() " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::delete_device) ENABLED START -----*/

	//	Delete device allocated objects
	//  Delete device's allocated object
  if(this->_electrometerCH1)
  {
    delete this->_electrometerCH1;
    this->_electrometerCH1 = 0;
  }

  if(this->_electrometerCH2)
  {
    delete this->_electrometerCH2;
    this->_electrometerCH2 = 0;
  }

  DELETE_DEVSTRING_ATTRIBUTE(attr_range1_read);
  DELETE_DEVSTRING_ATTRIBUTE(attr_range2_read);
  DELETE_SCALAR_ATTRIBUTE(attr_electrometerChannel_read);

  //- release the inner-appender
  INFO_STREAM << "Remove the inner-appender." << endl;
  yat4tango::InnerAppender::release(this);

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::delete_device
}

//--------------------------------------------------------
/**
 *	Method      : MCCE2Electrometers::init_device()
 *	Description : will be called at device initialization.
 */
//--------------------------------------------------------
void MCCE2Electrometers::init_device()
{
	DEBUG_STREAM << "MCCE2Electrometers::init_device() create device " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::init_device_before) ENABLED START -----*/

	//	Initialization before get_device_property() call

  // Initialize the inner appender.
  yat4tango::InnerAppender::initialize(this);
  //--------------------------------------------
  this->_electrometer    = 0;
  this->_electrometerCH1 = 0;
  this->_electrometerCH2 = 0;
  this->_missing_property = false;
  attr_electrometerChannel_write = ELETRO_CHANNEL_ONE;
  this->_updateRange1FromHW = true;    //- force update after init.
  this->_updateRange2FromHW = true;    //- force update after init.
  this->_electroCH1StrType  = "Not connected";
  this->_electroCH2StrType  = "Not connected";
  _notConnectedStr = "Not connected";



	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::init_device_before


	//	Get the device properties from database
	get_device_property();


	/*----- PROTECTED REGION ID(MCCE2Electrometers::init_device) ENABLED START -----*/

	//	Initialize device

  if ( this->proxyName.empty() )
    this->_missing_property = true;
  if ( !electrometer1Address && !electrometer2Address )
    this->_missing_property = true;
  INFO_STREAM << "electrometerTypeChx -> _missing_property = " << _missing_property << std::endl;

  CREATE_SCALAR_ATTRIBUTE(attr_electrometerChannel_read);
  CREATE_DEVSTRING_ATTRIBUTE(attr_range1_read, MAX_STRING_LENGTH, "Not updated yet.");
  CREATE_DEVSTRING_ATTRIBUTE(attr_range2_read, MAX_STRING_LENGTH, "Not updated yet.");

  this->_init_done = false;

  if( !this->_init_done && !this->_missing_property )
  {
    INFO_STREAM << "init_device -> create_electrometers_obj : ... " << std::endl;
    try
    {
      //- check connected electrometer(s)
      create_electrometers_obj();
    }
    catch(Tango::DevFailed& df)
    {
        FATAL_STREAM << "Init_dev : create_electrometers_obj -> DF :\n" << df << std::endl;
    }
    catch(...)
    {
      FATAL_STREAM << "Device initialization Failed : check logs for more informations.";

      set_state(Tango::FAULT);
      set_status("Failed to find which Novelec type is connected : communication caught[...]");
      return;
    }
  }//- end if

  INFO_STREAM << "Device up but not completly initialised!" << std::endl;
  INFO_STREAM << "this->proxyName = " << this->proxyName << std::endl;
  INFO_STREAM << "electrometer1Address = " << electrometer1Address << std::endl;
  INFO_STREAM << "electrometer2Address = " << electrometer2Address << std::endl;

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::init_device
}

//--------------------------------------------------------
/**
 *	Method      : MCCE2Electrometers::get_device_property()
 *	Description : Read database to initialize property data members.
 */
//--------------------------------------------------------
void MCCE2Electrometers::get_device_property()
{
	/*----- PROTECTED REGION ID(MCCE2Electrometers::get_device_property_before) ENABLED START -----*/

	//	Initialize property data members
	proxyName = "";
  electrometer1Address  = 0;
  electrometer2Address  = 0;

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::get_device_property_before


	//	Read device properties from database.
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("ProxyName"));
	dev_prop.push_back(Tango::DbDatum("Electrometer1Address"));
	dev_prop.push_back(Tango::DbDatum("Electrometer2Address"));
	dev_prop.push_back(Tango::DbDatum("ComProtocol"));

	//	is there at least one property to be read ?
	if (dev_prop.size()>0)
	{
		//	Call database and extract values
		if (Tango::Util::instance()->_UseDb==true)
			get_db_device()->get_property(dev_prop);

		//	get instance on MCCE2ElectrometersClass to get class property
		Tango::DbDatum	def_prop, cl_prop;
		MCCE2ElectrometersClass	*ds_class =
			(static_cast<MCCE2ElectrometersClass *>(get_device_class()));
		int	i = -1;

		//	Try to initialize ProxyName from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  proxyName;
		else {
			//	Try to initialize ProxyName from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  proxyName;
		}
		//	And try to extract ProxyName value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  proxyName;

		//	Try to initialize Electrometer1Address from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  electrometer1Address;
		else {
			//	Try to initialize Electrometer1Address from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  electrometer1Address;
		}
		//	And try to extract Electrometer1Address value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  electrometer1Address;

		//	Try to initialize Electrometer2Address from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  electrometer2Address;
		else {
			//	Try to initialize Electrometer2Address from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  electrometer2Address;
		}
		//	And try to extract Electrometer2Address value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  electrometer2Address;

		//	Try to initialize ComProtocol from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  comProtocol;
		else {
			//	Try to initialize ComProtocol from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  comProtocol;
		}
		//	And try to extract ComProtocol value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  comProtocol;

	}

	/*----- PROTECTED REGION ID(MCCE2Electrometers::get_device_property_after) ENABLED START -----*/

	//	Check device property data members init
	//- Creates default property
  Tango::DbData data_put;
  if ( dev_prop[0].is_empty() )
  {
    Tango::DbDatum  prop("ComProtocol");
    prop  <<  comProtocol;
    data_put.push_back(prop);
  }
  if ( dev_prop[0].is_empty() )
  {
    Tango::DbDatum  prop("ProxyName");
    prop  <<  proxyName;
    data_put.push_back(prop);
  }
  if ( dev_prop[1].is_empty() )
  {
    Tango::DbDatum  prop("Electrometer1Address");
    prop  <<  electrometer1Address;
    data_put.push_back(prop);
  }
  if ( dev_prop[2].is_empty() )
  {
    Tango::DbDatum  prop("Electrometer2Address");
    prop  <<  electrometer2Address;
    data_put.push_back(prop);
  }

  //- write default property if created
  if( !data_put.empty() )
    get_db_device()->put_property(data_put);

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::get_device_property_after
}

//--------------------------------------------------------
/**
 *	Method      : MCCE2Electrometers::always_executed_hook()
 *	Description : method always executed before any command is executed
 */
//--------------------------------------------------------
void MCCE2Electrometers::always_executed_hook()
{
	//INFO_STREAM << "MCCE2Electrometers::always_executed_hook()  " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::always_executed_hook) ENABLED START -----*/

	//	code always executed before all requests
	//- noop

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::always_executed_hook
}

//--------------------------------------------------------
/**
 *	Method      : MCCE2Electrometers::read_attr_hardware()
 *	Description : Hardware acquisition for attributes
 */
//--------------------------------------------------------
void MCCE2Electrometers::read_attr_hardware(TANGO_UNUSED(vector<long> &attr_list))
{
	//DEBUG_STREAM << "MCCE2Electrometers::read_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::read_attr_hardware) ENABLED START -----*/

	//	Add your own code

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::read_attr_hardware
}

//--------------------------------------------------------
/**
 *	Read attribute electrometerChannel related method
 *	Description: MCCE2 electrometer have two channels to support two amplifiers.\nWARN : default is 0 (please set this value first).
 *
 *	Data type:	Tango::DevUShort
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void MCCE2Electrometers::read_electrometerChannel(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MCCE2Electrometers::read_electrometerChannel(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::read_electrometerChannel) ENABLED START -----*/
	Tango::DevUShort channelNumber = 0;

  if ( !this->_electrometer )
  {
    ERROR_STREAM << "Failed to build the electrometer obj : error on connection or check_electrotype at init_device step?" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : error on connection or check_electrotype at init_device step",
      "MCCE2Electrometers::read_electrometerChannel()"
    );
  }

  try
  {
    *attr_electrometerChannel_read = this->_electrometer->get_ElectroChannel();
    attr.set_value(attr_electrometerChannel_read);
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    set_state(Tango::ALARM);
    set_status("Failed to set the active channel value!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to get the active channel value : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to get the active channel value.",
                                      "MCCE2Electrometers::read_electrometerChannel()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Failed to set the active channel value!");

    ERROR_STREAM << "Failed to get the active channel value : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to get the new active channel value!",
                                      "MCCE2Electrometers::read_electrometerChannel()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to get the active channel value : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to set the active channel value!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::read_electrometerChannel
}
//--------------------------------------------------------
/**
 *	Write attribute electrometerChannel related method
 *	Description: MCCE2 electrometer have two channels to support two amplifiers.\nWARN : default is 0 (please set this value first).
 *
 *	Data type:	Tango::DevUShort
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void MCCE2Electrometers::write_electrometerChannel(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MCCE2Electrometers::write_electrometerChannel(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevUShort	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(MCCE2Electrometers::write_electrometerChannel) ENABLED START -----*/

  if ( w_val == ELETRO_CHANNEL_ONE )
  {
    if ( this->_electrometerCH1 )
    {
      this->_electrometer = this->_electrometerCH1;
      //- now force update from HW
      this->_updateRange1FromHW = true;
      
      *attr_range1_read = get_range();
      
    }
    else
    {
      WARN_STREAM << "Cannot control electrometer on this channel : seems that nothing is connected on!" << std::endl;
      // nothing is connected on this channel      
      *attr_range1_read = &_notConnectedStr[0];
    }
  }
  else if ( w_val == ELETRO_CHANNEL_TWO )
  {
    if ( this->_electrometerCH2 )
    {
      this->_electrometer = this->_electrometerCH2;
      //- now force update from HW
      this->_updateRange2FromHW = true;
      
      *attr_range2_read = get_range();
      
    }
    else
    {
      WARN_STREAM << "Cannot control electrometer on this channel : seems that nothing is connected on!" << std::endl;
      //- warn nothing is connected on this channel
      *attr_range2_read = &_notConnectedStr[0];
    }
  }
  attr_electrometerChannel_write = w_val;

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::write_electrometerChannel
}
//--------------------------------------------------------
/**
 *	Read attribute range1 related method
 *	Description: The channel one MCCE2 range.
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar */
//--------------------------------------------------------
void MCCE2Electrometers::read_range1(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MCCE2Electrometers::read_range1(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::read_range1) ENABLED START -----*/
	if ( !this->_electrometerCH1 )
  {
    *attr_range1_read = &_notConnectedStr[0];
  }
  else
  {
    if ( this->_updateRange1FromHW && *attr_electrometerChannel_read == ELETRO_CHANNEL_ONE )
    {
      *attr_range1_read = this->get_range();
      
      this->_updateRange1FromHW = false;
    }
  }
  attr.set_value(attr_range1_read);

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::read_range1
}
//--------------------------------------------------------
/**
 *	Read attribute range2 related method
 *	Description: The channel two MCCE range.
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void MCCE2Electrometers::read_range2(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MCCE2Electrometers::read_range2(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::read_range2) ENABLED START -----*/
	if ( !this->_electrometerCH2 )
  {
    *attr_range2_read = &_notConnectedStr[0];
  }
  else
  {
    if ( this->_updateRange2FromHW && *attr_electrometerChannel_read == ELETRO_CHANNEL_TWO )
    {
     *attr_range2_read = get_range();
      
      this->_updateRange2FromHW = false;
    }
  }
  attr.set_value(attr_range2_read);

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::read_range2
}

//--------------------------------------------------------
/**
 *	Method      : MCCE2Electrometers::add_dynamic_attributes()
 *	Description : Create the dynamic attributes if any
 *                for specified device.
 */
//--------------------------------------------------------
void MCCE2Electrometers::add_dynamic_attributes()
{
	/*----- PROTECTED REGION ID(MCCE2Electrometers::add_dynamic_attributes) ENABLED START -----*/

	//	Add your own code to create and add dynamic attributes if any

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::add_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Command State related method
 *	Description: This command gets the device state (stored in its <i>device_state</i> data member) and returns it to the caller.
 *
 *	@returns State Code
 */
//--------------------------------------------------------
Tango::DevState MCCE2Electrometers::dev_state()
{
	DEBUG_STREAM << "MCCE2Electrometers::State()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::dev_state) ENABLED START -----*/

	Tango::DevState	argout = Tango::UNKNOWN; // replace by your own algorithm
	//	Add your own code

  try
  {
    //- check properties
    if ( this->_missing_property )
    {
      argout = Tango::FAULT;
    }
    //- check device initialization
    else  if ( !this->_electrometerCH1 && !this->_electrometerCH2 )
    {
      argout = Tango::FAULT;
    }
    //- check if an electro is conencted on a channel
    else if( attr_electrometerChannel_write == ELETRO_CHANNEL_ONE && !this->_electrometerCH1)
    {
      argout = Tango::DISABLE;
    }
    else if( attr_electrometerChannel_write == ELETRO_CHANNEL_TWO && !this->_electrometerCH2)
    {
      argout = Tango::DISABLE;
    }
    //- check channel number
    else if ( attr_electrometerChannel_write != ELETRO_CHANNEL_ONE && attr_electrometerChannel_write != ELETRO_CHANNEL_TWO )
    {
      argout = Tango::ALARM;
    }
    else
    {
      //- update state
      _electrometer->electrometer_status();

      switch(_electrometer->electrometer_state())
      {
      case Tango::ON    :
        argout = Tango::ON;
        break;
      case Tango::FAULT :
        argout = Tango::FAULT;
        break;
      case Tango::ALARM  :
        argout = Tango::ALARM;
        break;
        /*case Tango::UNKNOWN:
        default :
            argout = Tango::UNKNOWN;
            break;*/
      }
    }
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    set_state(Tango::ALARM);
    set_status("Failed to get electrometer state!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Cannot perform get electrometer state : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to get electrometer state.",
                                      "MCCE2Electrometers::dev_state()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Failed to get electrometer state!");

    ERROR_STREAM << "Failed to get electrometer state : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to get electrometer state!",
                                      "MCCE2Electrometers::dev_state()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    set_state(Tango::ALARM);
    set_status("Failed to get electrometer state!");
    ERROR_STREAM << "Failed to get electrometer state : [...] EXCEPTION " << std::endl;
    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::dev_state
	set_state(argout);    // Give the state to Tango.
	if (argout!=Tango::ALARM)
		DeviceImpl::dev_state();
	return get_state();  // Return it after Tango management.
}
//--------------------------------------------------------
/**
 *	Command Status related method
 *	Description: This command gets the device status (stored in its <i>device_status</i> data member) and returns it to the caller.
 *
 *	@returns Status description
 */
//--------------------------------------------------------
Tango::ConstDevString MCCE2Electrometers::dev_status()
{
	DEBUG_STREAM << "MCCE2Electrometers::Status()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::dev_status) ENABLED START -----*/

	string	status = "Device is OK";
	//	Add your own code
	//- clear previous status
  _statusStr.clear();

  //  Add your own code to control device here
  try
  {
    if ( this->_missing_property )
    {
      _statusStr = "Device cannot be up while its properties are not set!";
    }

    else if ( !this->_electrometer )
    {
    	_statusStr = " Init device failed: communication error or something wrong on check_electrotype";
    }
    //- check device initialization
    else  if ( !this->_electrometerCH1 && !_electrometerCH2 )
    {
      _statusStr = "No electrometer connected!";
    }

    //- check if an electro is connected on a channel
    if( attr_electrometerChannel_write == ELETRO_CHANNEL_ONE && !this->_electrometerCH1)
    {
      _statusStr += "No electrometer on channel ONE!";
    }
    else if( attr_electrometerChannel_write == ELETRO_CHANNEL_TWO && !this->_electrometerCH2)
    {
      _statusStr += "No electrometer on channel TWO!";
    }
    else if ( attr_electrometerChannel_write != ELETRO_CHANNEL_ONE && attr_electrometerChannel_write != ELETRO_CHANNEL_TWO )
    {
      _statusStr += "Invalid electrometer channel set!";
    }
    else
    {
      //- add electrometer status info
      _statusStr += this->_electrometer->electrometer_status();
    }
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    set_state(Tango::ALARM);
    set_status("Cannot perform electrometer_status() command.");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Cannot perform electrometer_status() command : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to get electrometer status.",
                                      "MCCE2Electrometers::dev_status()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Cannot perform electrometer_status() command!");

    ERROR_STREAM << "Cannot perform electrometer_status() command : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Cannot perform electrometer_status() command!",
                                      "MCCE2Electrometers::dev_status()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to enable remote mode : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Cannot perform electrometer_status() command : caugth [...] exception.");
    throw;
  }
  status = _statusStr;

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::dev_status
	set_status(status);               // Give the status to Tango.
	return DeviceImpl::dev_status();  // Return it.
}
//--------------------------------------------------------
/**
 *	Command RangeUP related method
 *	Description:
 *
 */
//--------------------------------------------------------
void MCCE2Electrometers::range_up()
{
	DEBUG_STREAM << "MCCE2Electrometers::RangeUP()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::range_up) ENABLED START -----*/

	//	Add your own code
	if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::range_up()"
    );
  }

  try
  {
    this->_electrometer->range_up();
    //- update range attr from hardware
    if ( *attr_electrometerChannel_read == ELETRO_CHANNEL_ONE)
      this->_updateRange1FromHW = true;
    else if ( *attr_electrometerChannel_read == ELETRO_CHANNEL_TWO)
      this->_updateRange2FromHW = true;
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    set_state(Tango::ALARM);
    set_status("Failed to up the range!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to up the range : DevFailed EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to up the range.",
                                      "MCCE2Electrometers::range_up()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Failed to up the range!");

    ERROR_STREAM << "Failed to up the range : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to up the range!",
                                      "MCCE2Electrometers::range_up()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to up the range : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to up the range!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::range_up
}
//--------------------------------------------------------
/**
 *	Command RangeDOWN related method
 *	Description:
 *
 */
//--------------------------------------------------------
void MCCE2Electrometers::range_down()
{
	DEBUG_STREAM << "MCCE2Electrometers::RangeDOWN()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::range_down) ENABLED START -----*/

	//	Add your own code
	if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::range_down()"
    );
  }

  try
  {
    this->_electrometer->range_down();
    //- update range attr from hardware
    if ( *attr_electrometerChannel_read == ELETRO_CHANNEL_ONE )
      this->_updateRange1FromHW = true;
    else if ( *attr_electrometerChannel_read == ELETRO_CHANNEL_TWO )
      this->_updateRange2FromHW = true;
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    set_state(Tango::ALARM);
    set_status("Failed to down the range!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to down the range : DevFailed EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to down the range.",
                                      "MCCE2Electrometers::range_down()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Failed to down the range!");

    ERROR_STREAM << "Failed to down the range : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to down the range!",
                                      "MCCE2Electrometers::range_down()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to down the range : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to down the range!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::range_down
}
//--------------------------------------------------------
/**
 *	Command SetZeroVFFunction related method
 *	Description:
 *
 */
//--------------------------------------------------------
void MCCE2Electrometers::set_zero_vffunction()
{
	DEBUG_STREAM << "MCCE2Electrometers::SetZeroVFFunction()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::set_zero_vffunction) ENABLED START -----*/

	//	Add your own code
	if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::set_zero_vffunction()"
    );
  }

  try
  {
    this->_electrometer->set_Zero_VonF_function();
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable ZERO V/F function!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to enable ZERO V/F function : DevFailed EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to enable ZERO V/F function.",
                                      "MCCE2Electrometers::set_zero_vffunction()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable ZERO V/F function!");

    ERROR_STREAM << "Failed to enable ZERO V/F function : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to enable ZERO V/F function!",
                                      "MCCE2Electrometers::set_zero_vffunction()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to enable ZERO V/F function : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to enable ZERO V/F function!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::set_zero_vffunction
}
//--------------------------------------------------------
/**
 *	Command SetOffsetZeroV1Function related method
 *	Description:
 *
 */
//--------------------------------------------------------
void MCCE2Electrometers::set_offset_zero_v1_function()
{
	DEBUG_STREAM << "MCCE2Electrometers::SetOffsetZeroV1Function()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::set_offset_zero_v1_function) ENABLED START -----*/

	//	Add your own code
	if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::set_offset_zero_v1_function()"
    );
  }

  try
  {
    this->_electrometer->set_Offset_ZeroV1_function();
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable OFFSET or ZERO V1 function!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to enable OFFSET or ZERO V1 function : DevFailed EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to enable OFFSET or ZERO V1 function.",
                                      "MCCE2Electrometers::set_offset_zero_v1_function()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable OFFSET or ZERO V1 function!");

    ERROR_STREAM << "Failed to enable OFFSET or ZERO V1 function : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to enable OFFSET or ZERO V1 function!",
                                      "MCCE2Electrometers::set_offset_zero_v1_function()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to enable OFFSET or ZERO V1 function : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to enable OFFSET or ZERO V1 function!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::set_offset_zero_v1_function
}
//--------------------------------------------------------
/**
 *	Command SetLeakageZeroV2Function related method
 *	Description:
 *
 */
//--------------------------------------------------------
void MCCE2Electrometers::set_leakage_zero_v2_function()
{
	DEBUG_STREAM << "MCCE2Electrometers::SetLeakageZeroV2Function()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::set_leakage_zero_v2_function) ENABLED START -----*/

	//	Add your own code
	if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::set_leakage_zero_v2_function()"
    );
  }

  try
  {
    this->_electrometer->set_Leakage_ZeroV2_function();
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable LEAKAGE or ZERO V2 function!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to enable LEAKAGE or ZERO V2 function : DevFailed EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to enable LEAKAGE or ZERO V2 function.",
                                      "MCCE2Electrometers::set_leakage_zero_v2_function()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable LEAKAGE or ZERO V2 function!");

    ERROR_STREAM << "Failed to enable LEAKAGE or ZERO V2 function : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to enable LEAKAGE or ZERO V2 function!",
                                      "MCCE2Electrometers::set_leakage_zero_v2_function()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to enable LEAKAGE or ZERO V2 function : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to enable LEAKAGE or ZERO V2 function!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::set_leakage_zero_v2_function
}
//--------------------------------------------------------
/**
 *	Command SetTestFunction related method
 *	Description:
 *
 */
//--------------------------------------------------------
void MCCE2Electrometers::set_test_function()
{
	DEBUG_STREAM << "MCCE2Electrometers::SetTestFunction()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::set_test_function) ENABLED START -----*/

	//	Add your own code
	if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::set_test_function()"
    );
  }

  try
  {
    this->_electrometer->set_Test_function();
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable TEST function!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to enable TEST function : DevFailed EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to enable TEST function.",
                                      "MCCE2Electrometers::set_test_function()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable TEST function!");

    ERROR_STREAM << "Failed to enable TEST function : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to enable TEST function!",
                                      "MCCE2Electrometers::set_test_function()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to enable TEST function : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to enable TEST function!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::set_test_function
}
//--------------------------------------------------------
/**
 *	Command SetMeasureFunction related method
 *	Description:
 *
 */
//--------------------------------------------------------
void MCCE2Electrometers::set_measure_function()
{
	DEBUG_STREAM << "MCCE2Electrometers::SetMeasureFunction()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::set_measure_function) ENABLED START -----*/

	//	Add your own code
	if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::set_measure_function()"
    );
  }

  try
  {
    this->_electrometer->set_Measure_function();
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    Tango::DevFailed df = electrometer_to_tango_exception(ke);
    set_state(Tango::ALARM);
    set_status("Failed to enable MEASURE function!");

    ERROR_STREAM << "Failed to enable MEASURE function : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to enable MEASURE function.",
                                      "MCCE2Electrometers::set_measure_function()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable MEASURE function!");

    ERROR_STREAM << "Failed to enable MEASURE function : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to enable MEASURE function!",
                                      "MCCE2Electrometers::set_measure_function()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to enable MEASURE function : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to enable MEASURE function!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::set_measure_function
}
//--------------------------------------------------------
/**
 *	Command MCCE2_ON related method
 *	Description:
 *
 */
//--------------------------------------------------------
void MCCE2Electrometers::mcce2__on()
{
	DEBUG_STREAM << "MCCE2Electrometers::MCCE2_ON()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::mcce2__on) ENABLED START -----*/

	//	Add your own code
	if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::mcce2__on()"
    );
  }

  try
  {
    this->_electrometer->mcce_on();
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable MCCE ON function!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to enable MCCE ON function : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to enable MCCE ON function.",
                                      "MCCE2Electrometers::mcce2__on()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable MCCE ON function!");

    ERROR_STREAM << "Failed to enable MCCE ON function : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to enable MCCE ON function!",
                                      "MCCE2Electrometers::mcce2__on()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to enable MCCE2_ON function : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to enable MCCE ON function!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::mcce2__on
}
//--------------------------------------------------------
/**
 *	Command MCCE2_OFF related method
 *	Description:
 *
 */
//--------------------------------------------------------
void MCCE2Electrometers::mcce2__off()
{
	DEBUG_STREAM << "MCCE2Electrometers::MCCE2_OFF()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::mcce2__off) ENABLED START -----*/

	//	Add your own code
	if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::mcce2__off()"
    );
  }

  try
  {
    this->_electrometer->mcce_off();
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable MCCE OFF function!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to enable MCCE OFF function : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to enable MCCE OFF function.",
                                      "MCCE2Electrometers::mcce2__off()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable MCCE OFF function!");

    ERROR_STREAM << "Failed to enable MCCE OFF function : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to enable MCCE OFF function!",
                                      "MCCE2Electrometers::mcce2__off()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to enable MCCE2_OFF function : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to enable MCCE OFF function!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::mcce2__off
}
//--------------------------------------------------------
/**
 *	Command Local related method
 *	Description:
 *
 */
//--------------------------------------------------------
void MCCE2Electrometers::local()
{
	DEBUG_STREAM << "MCCE2Electrometers::Local()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::local) ENABLED START -----*/

	//	Add your own code
	if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::local()"
    );
  }

  try
  {
    this->_electrometer->local();
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable local mode!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to enable local mode : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to enable local mode.",
                                      "MCCE2Electrometers::local()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable local mode!");

    ERROR_STREAM << "Failed to enable local function : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to enable local mode!",
                                      "MCCE2Electrometers::local()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to enable local mode : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to enable local mode!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::local
}
//--------------------------------------------------------
/**
 *	Command Remote related method
 *	Description:
 *
 */
//--------------------------------------------------------
void MCCE2Electrometers::remote()
{
	DEBUG_STREAM << "MCCE2Electrometers::Remote()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::remote) ENABLED START -----*/

	//	Add your own code
	if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::remote()"
    );
  }

  try
  {
    this->_electrometer->remote();
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable remote mode!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to enable remote mode : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to enable remote mode.",
                                      "MCCE2Electrometers::remote()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable remote mode!");

    ERROR_STREAM << "Failed to enable remote mode : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to enable remote mode!",
                                      "MCCE2Electrometers::remote()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to enable remote mode : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to enable remote mode!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::remote
}
//--------------------------------------------------------
/**
 *	Command Reset related method
 *	Description:
 *
 */
//--------------------------------------------------------
void MCCE2Electrometers::reset()
{
	DEBUG_STREAM << "MCCE2Electrometers::Reset()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::reset) ENABLED START -----*/

	//	Add your own code
	if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::reset()"
    );
  }

  try
  {
    this->_electrometer->reset();
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable reset mode!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to enable reset mode : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to enable reset mode.",
                                      "MCCE2Electrometers::reset()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Failed to enable reset mode!");

    ERROR_STREAM << "Failed to enable reset mode : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to enable reset mode!",
                                      "MCCE2Electrometers::reset()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to enable reset mode : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to enable reset mode!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::reset
}
//--------------------------------------------------------
/**
 *	Command ClearRegisters related method
 *	Description:
 *
 */
//--------------------------------------------------------
void MCCE2Electrometers::clear_registers()
{
	DEBUG_STREAM << "MCCE2Electrometers::ClearRegisters()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::clear_registers) ENABLED START -----*/

	//	Add your own code
	if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::clear_registers()"
    );
  }

  try
  {
    this->_electrometer->clear_registers();
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    set_state(Tango::ALARM);
    set_status("Failed to clear all errors!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to clear all errors : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to clear all errors.",
                                      "MCCE2Electrometers::clear_registers()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Failed to clear all errors!");

    ERROR_STREAM << "Failed to clear all errors : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to clear all errors!",
                                      "MCCE2Electrometers::clear_registers()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to clear errors : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to clear all errors!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::clear_registers
}
//--------------------------------------------------------
/**
 *	Command GetRange related method
 *	Description:
 *
 *	@returns The actual electrometer range
 */
//--------------------------------------------------------
Tango::DevString MCCE2Electrometers::get_range()
{
	Tango::DevString argout;
	DEBUG_STREAM << "MCCE2Electrometers::GetRange()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::get_range) ENABLED START -----*/

  //  Add your own code to control device here
  if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::get_range()"
    );
  }

  try
  {
    std::string rangeStr("");
    rangeStr = this->_electrometer->get_ElectroMeterRange();
    argout = new char [rangeStr.size() + 1];
    strcpy(argout, rangeStr.c_str());
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    if ( argout )
      delete [] argout;
    Tango::DevFailed df = electrometer_to_tango_exception(ke);
    set_state(Tango::ALARM);
    set_status("Failed to read the range!");

    ERROR_STREAM << "Failed to read the range : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to read the range.",
                                      "MCCE2Electrometers::get_range()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    if ( argout )
      delete [] argout;
    set_state(Tango::ALARM);
    set_status("Failed to read the range!");

    ERROR_STREAM << "Failed to read the range : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to read the range!",
                                      "MCCE2Electrometers::get_range()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    if ( argout )
      delete [] argout;
    ERROR_STREAM << "Failed to read the range : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to read the range!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::get_range
	return argout;
}
//--------------------------------------------------------
/**
 *	Command SetRange related method
 *	Description:
 *
 *	@param argin The range to apply on the electrometer
 */
//--------------------------------------------------------
void MCCE2Electrometers::set_range(Tango::DevString argin)
{
	DEBUG_STREAM << "MCCE2Electrometers::SetRange()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::set_range) ENABLED START -----*/

	//	Add your own code
	if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::set_range()"
    );
  }

  try
  {
    this->_electrometer->set_ElectroMeterRange(std::string(argin));
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    set_state(Tango::ALARM);
    set_status("Failed to write the range!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to write the range : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to write the range.",
                                      "MCCE2Electrometers::set_range()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::ALARM);
    set_status("Failed to write the range!");

    ERROR_STREAM << "Failed to write the range : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to write the range!",
                                      "MCCE2Electrometers::set_range()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to set the new range value : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to write the range!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::set_range
}
//--------------------------------------------------------
/**
 *	Command GetMode related method
 *	Description:
 *
 *	@returns The electrometer mode (MEASURE, LEAKAGE ...)
 */
//--------------------------------------------------------
Tango::DevString MCCE2Electrometers::get_mode()
{
	Tango::DevString argout;
	DEBUG_STREAM << "MCCE2Electrometers::GetMode()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::get_mode) ENABLED START -----*/

	//	Add your own code
	//  POGO has generated a method core with argout allocation.
  //  If you would like to use a static reference without copying,
  //  See "TANGO Device Server Programmer's Manual"
  //    (chapter : Writing a TANGO DS / Exchanging data)
  //------------------------------------------------------------
  DEBUG_STREAM << "MCCE2Electrometers::get_mode(): entering... !" << endl;

  //  Add your own code to control device here
  if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::get_mode()"
    );
  }


  try
  {
    std::string modeStr = this->_electrometer->get_ElectroMeterMode();
    argout = new char [modeStr.size() + 1];
    strcpy(argout, modeStr.c_str());
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    if ( argout )
      delete [] argout;
    set_state(Tango::ALARM);
    set_status("Failed to read the mode!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to set read the mode : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to read the mode.",
                                      "MCCE2Electrometers::get_mode()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    if ( argout )
      delete [] argout;
    set_state(Tango::ALARM);
    set_status("Failed to read the mode!");

    ERROR_STREAM << "Failed to read the mode : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to read the mode!",
                                      "MCCE2Electrometers::get_mode()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    if ( argout )
      delete [] argout;
    ERROR_STREAM << "Failed to read the mode : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to read the mode!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::get_mode
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetPolarity related method
 *	Description:
 *
 *	@returns The electrometer polarity
 */
//--------------------------------------------------------
Tango::DevString MCCE2Electrometers::get_polarity()
{
	Tango::DevString argout;
	DEBUG_STREAM << "MCCE2Electrometers::GetPolarity()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::get_polarity) ENABLED START -----*/

	//	Add your own code
	//  POGO has generated a method core with argout allocation.
  //  If you would like to use a static reference without copying,
  //  See "TANGO Device Server Programmer's Manual"
  //    (chapter : Writing a TANGO DS / Exchanging data)
  //------------------------------------------------------------
  DEBUG_STREAM << "MCCE2Electrometers::get_polarity(): entering... !" << endl;

  //  Add your own code to control device here
  if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::get_polarity()"
    );
  }

  try
  {
    std::string polarStr = this->_electrometer->get_ElectroMeterPolarity();
    argout = new char [polarStr.size() + 1];
    strcpy(argout, polarStr.c_str());
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    if ( argout )
      delete [] argout;
    set_state(Tango::ALARM);
    set_status("Failed to read the polarity!");
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to set read the polarity : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to read the polarity.",
                                      "MCCE2Electrometers::get_polarity()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    if ( argout )
      delete [] argout;
    set_state(Tango::ALARM);
    set_status("Failed to read the polarity!");

    ERROR_STREAM << "Failed to read the polarity : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to read the polarity!",
                                      "MCCE2Electrometers::get_polarity()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    if ( argout )
      delete [] argout;
    ERROR_STREAM << "Failed to read the polarity : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to read the polarity!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::get_polarity
	return argout;
}
//--------------------------------------------------------
/**
 *	Command SetPolarity related method
 *	Description:
 *
 *	@param argin POSITIVE or NEGATIVE
 */
//--------------------------------------------------------
void MCCE2Electrometers::set_polarity(Tango::DevString argin)
{
	DEBUG_STREAM << "MCCE2Electrometers::SetPolarity()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::set_polarity) ENABLED START -----*/

	//	Add your own code
	if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::set_polarity()"
    );
  }

  try
  {
    this->_electrometer->set_ElectroMeterPolarity(argin);
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    set_state(Tango::ALARM);
    set_status("Failed to change electrometer polarity!");
    ERROR_STREAM << "Failed to change electrometer polarity : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to change electrometer polarity.",
                                      "hzFilterAttrib::set_polarity()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    ERROR_STREAM << "Failed to change electrometer polarity : DevFailed EXCEPTION -> " << df << std::endl;

    set_state(Tango::ALARM);
    set_status("Failed to change electrometer polarity!");
    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to change electrometer polarity!",
                                      "MCCE2Electrometers::set_polarity()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to change electrometer polarity : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to change electrometer polarity!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::set_polarity
}
//--------------------------------------------------------
/**
 *	Command GetHzFilter related method
 *	Description:
 *
 *	@returns The MCCE2 cut-off frequency in Hz
 */
//--------------------------------------------------------
Tango::DevShort MCCE2Electrometers::get_hz_filter()
{
	Tango::DevShort argout;
	DEBUG_STREAM << "MCCE2Electrometers::GetHzFilter()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::get_hz_filter) ENABLED START -----*/

	//	Add your own code
  std::string dataStr("");

  if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::get_hz_filter()"
    );
  }

  try
  {
    dataStr = this->_electrometer->get_ElectroMeterFrequency();
    argout  = yat::XString<short>::to_num(dataStr);
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    set_state(Tango::ALARM);
    set_status("Failed to set the Hz Filter!");
    ERROR_STREAM << "Failed to read the gain : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to read hzFilter.",
                                      "hzFilterAttrib::get_hz_filter()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    ERROR_STREAM << "Failed to read the Hz Filter : DevFailed EXCEPTION -> " << df << std::endl;

    set_state(Tango::ALARM);
    set_status("Failed to set the Hz Filter!");
    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to read the Hz Filter!",
                                      "MCCE2Electrometers::get_hz_filter()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to read the Hz Filter : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to read the Hz Filter!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::get_hz_filter
	return argout;
}
//--------------------------------------------------------
/**
 *	Command SetHzFilter related method
 *	Description:
 *
 *	@param argin The MCCE2 cut-off frequency in Hz
 */
//--------------------------------------------------------
void MCCE2Electrometers::set_hz_filter(Tango::DevShort argin)
{
	DEBUG_STREAM << "MCCE2Electrometers::SetHzFilter()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::set_hz_filter) ENABLED START -----*/

	//	Add your own code
	std::string dataStr("") ;

  //  Add your own code to control device here
  if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::set_hz_filter()"
    );
  }

  try
  {
    dataStr = yat::XString<short>::to_string(argin);
    this->_electrometer->set_ElectroMeterFrequency(dataStr);
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    set_state(Tango::ALARM);
    set_status("Failed to set the Hz Filter!");
    ERROR_STREAM << "Failed to set the Hz Filter : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to set the Hz Filter.",
                                      "MCCE2Electrometers::set_hz_filter()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    ERROR_STREAM << "Failed to set the Hz Filter : DevFailed EXCEPTION -> " << df << std::endl;

    set_state(Tango::ALARM);
    set_status("Failed to set the Hz Filter!");
    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to set the Hz Filter!",
                                      "MCCE2Electrometers::set_hz_filter()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to set the Hz Filter : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to set the Hz Filter!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::set_hz_filter
}
//--------------------------------------------------------
/**
 *	Command GetGain related method
 *	Description:
 *
 *	@returns The MCCE2 gain
 */
//--------------------------------------------------------
Tango::DevShort MCCE2Electrometers::get_gain()
{
	Tango::DevShort argout;
	DEBUG_STREAM << "MCCE2Electrometers::GetGain()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::get_gain) ENABLED START -----*/

	//	Add your own code

  if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::get_gain()"
    );
  }

  try
  {
    std::string dataStr("");
    dataStr = this->_electrometer->get_ElectroMeterGain();
    argout  = yat::XString<short>::to_num(dataStr);
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    set_state(Tango::ALARM);
    set_status("Failed to read the gain!");
    ERROR_STREAM << "Failed to read the gain : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to read the gain .",
                                      "hzFilterAttrib::get_gain()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    ERROR_STREAM << "Failed to read the gain : DevFailed EXCEPTION -> " << df << std::endl;

    set_state(Tango::ALARM);
    set_status("Failed to read the gain!");
    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to read the gain!",
                                      "MCCE2Electrometers::get_gain()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to read the gain : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to read the gain!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::get_gain
	return argout;
}
//--------------------------------------------------------
/**
 *	Command SetGain related method
 *	Description:
 *
 *	@param argin The new MCCE-2 gain
 */
//--------------------------------------------------------
void MCCE2Electrometers::set_gain(Tango::DevUShort argin)
{
	DEBUG_STREAM << "MCCE2Electrometers::SetGain()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::set_gain) ENABLED START -----*/

	//	Add your own code
	if ( !this->_electrometer )
  {
    ERROR_STREAM << "Not enough memory to create the electrometer obj!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "Failed to build the electrometer obj : not enough memory!?",
      "MCCE2Electrometers::set_gain()"
    );
  }

  try
  {
    std::string dataStr("");
    dataStr = yat::XString<short>::to_string(argin);
    this->_electrometer->set_ElectroMeterGain(dataStr);
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    set_state(Tango::ALARM);
    set_status("Failed to set the gain!");
    ERROR_STREAM << "Failed to set the gain : ElectrometerException EXCEPTION -> " << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "UNKNOWN_ERROR",
                                      "Failed to set the gain.",
                                      "hzFilterAttrib::set_gain()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    ERROR_STREAM << "Failed to set the gain : DevFailed EXCEPTION -> " << df << std::endl;

    set_state(Tango::ALARM);
    set_status("Failed to set the gain!");
    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to set the the gain!",
                                      "MCCE2Electrometers::set_gain()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    ERROR_STREAM << "Failed to set the gain : [...] EXCEPTION " << std::endl;
    set_state(Tango::ALARM);
    set_status("Failed to set the gain!");

    throw;
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::set_gain
}
//--------------------------------------------------------
/**
 *	Command GetElectrometerTypeStr related method
 *	Description:
 *
 *	@param argin channel number
 *	@returns the electrometer type as string
 */
//--------------------------------------------------------
Tango::DevString MCCE2Electrometers::get_electrometer_type_str(Tango::DevUShort argin)
{
	Tango::DevString argout;
	DEBUG_STREAM << "MCCE2Electrometers::GetElectrometerTypeStr()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(MCCE2Electrometers::get_electrometer_type_str) ENABLED START -----*/


  if ( !this->_electrometerCH1 && !_electrometerCH2 )
  {
    ERROR_STREAM << "No electrometer connected on both channels!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "No electrometer connected on both channels!?",
      "MCCE2Electrometers::get_electrometer_type_str()"
    );
  }

  if ( argin == ELETRO_CHANNEL_ONE )
  {
    argout = new char[this->_electroCH1StrType.size()+1];
    ::strncpy(argout, _electroCH1StrType.c_str(), _electroCH1StrType.size()+1);
  }
  else if ( argin == ELETRO_CHANNEL_TWO )
  {
    argout = new char[this->_electroCH2StrType.size()+1];
    ::strncpy(argout, _electroCH2StrType.c_str(), _electroCH2StrType.size()+1);
  }
  else
  {
    Tango::Except::throw_exception(
      "INVALID_PARAMETER",
      "Channel number must be 1 or 2.",
      "MCCE2Electrometers::get_electrometer_type_str()"
    );
  }

	/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::get_electrometer_type_str
	return argout;
}

/*----- PROTECTED REGION ID(MCCE2Electrometers::namespace_ending) ENABLED START -----*/

//	Additional Methods
//+----------------------------------------------------------------------------
//
// method : MCCE2Electrometers::electrometer_to_tango_exception()
//
// description :   Extract real attribute values from
//        hardware acquisition result.
//
//-----------------------------------------------------------------------------
Tango::DevFailed MCCE2Electrometers::electrometer_to_tango_exception(const electrometer::ElectrometerException& de)
{
  Tango::DevErrorList error_list(de.errors.size());
  error_list.length(de.errors.size());

  for(unsigned int i = 0; i < de.errors.size(); i++)
  {
    error_list[i].reason = CORBA::string_dup(de.errors[i].reason.c_str());
    error_list[i].desc   = CORBA::string_dup(de.errors[i].desc.c_str());
    error_list[i].origin = CORBA::string_dup(de.errors[i].origin.c_str());

    switch(de.errors[i].severity)
    {
    case electrometer::WARN:
      error_list[i].severity = Tango::WARN;
      break;

    case electrometer::PANIC:
      error_list[i].severity = Tango::PANIC;
      break;

    case electrometer::ERR:

    default:
      error_list[i].severity = Tango::ERR;
      break;
    }
  }
  return Tango::DevFailed(error_list);
}
//+------------------------------------------------------------------
/**
 *  method: MCCE2Electrometers::create_electrometers_obj
 *
 *  description:  internal method to instanciate electrometer(s) obj
 */
//+------------------------------------------------------------------
void MCCE2Electrometers::create_electrometers_obj()
{
  //- instanciate channel 1
  if ( !_electrometerCH1 && electrometer1Address )
  {
    AbstractElectrometerClass* mcce1;
    mcce1 = new Novelec_MCCE2(proxyName, electrometer1Address, comProtocol);
    mcce1->init_protocol();
    DEBUG_STREAM<<"create_electrometers_obj init: "<< comProtocol <<std::endl;
    Novelec_MCCE2* nMCCE1 = dynamic_cast<Novelec_MCCE2*>(mcce1);
    DEBUG_STREAM<<"create_electrometers_obj init: "<< comProtocol <<std::endl;
    short type1 = 0;
    //- find connected Novelec model
    try
    {
      if(nMCCE1)
      {
        //- get type as num
        type1 = nMCCE1->get_electrometer_type();
        DEBUG_STREAM<<"get_electrometer_type type1= "<<type1 <<std::endl;
        //- get type as string
        this->_electroCH1StrType = nMCCE1->get_electrometer_typeStr();
        DEBUG_STREAM<<"get_electrometer_type:read "<<_electroCH1StrType <<std::endl;
      }
    }
    catch(const electrometer::ElectrometerException& ke)
    {
      set_state(Tango::ALARM);
      set_status("Failed to find which Novelec type is connected on channel one : communication caught[...]");

      Tango::DevFailed df = electrometer_to_tango_exception(ke);
      FATAL_STREAM << "Failed to find which Novelec type is connected on channel one : ElectrometerException EXCEPTION -> " << df << std::endl;
    }
    catch(Tango::DevFailed& df)
    {
      set_state(Tango::ALARM);
      set_status("Failed to find which Novelec type is connected on channel one!");

      FATAL_STREAM << "Failed to find which Novelec type is connected : DevFailed EXCEPTION -> " << df << std::endl;
    }
    catch(...)
    {
      set_state(Tango::ALARM);
      set_status("Failed to find which Novelec type is connected on channel one : communication caught[...]");

      FATAL_STREAM << "Failed to find which Novelec type is connected on channel one : [...] EXCEPTION " << std::endl;
    }

    switch(type1)
    {
    case 0 :
      this->_electrometerCH1 = 0;
      break;
    case 1 :
    case 2 :
    case 3 :
//            this->_electrometerCH1 = new N_PhotoVoltaique (proxyName, channelNumber, electrometerTypeCh1);
      this->_electrometerCH1 = new N_PhotoVoltaique (proxyName, electrometer1Address,comProtocol);
      break;
    case 4 :
    case 5 :
      this->_electrometerCH1 = new N_PhotoConducteur(proxyName, electrometer1Address,comProtocol);
      break;
    default :
      this->_electrometerCH1 = 0;
      FATAL_STREAM << "\nERROR : Electrometers::create_electrometers_obj() -> this " << electrometer1Address << " Novelec type is not supported now!" << endl;
      return;
    }
    if ( mcce1 )
    {
      delete mcce1;
      mcce1 = 0;
    }
  }

  //- instanciate channel 2
  if ( !this->_electrometerCH2 && electrometer2Address )
  {
    AbstractElectrometerClass* mcce2;
    mcce2 = new Novelec_MCCE2(proxyName, electrometer2Address,comProtocol);
    mcce2->init_protocol();
    Novelec_MCCE2* nMCCE2 = dynamic_cast<Novelec_MCCE2*>(mcce2);
    short type2 = 0;
    try
    {
      if(nMCCE2)
      {
        //- get type as num
        type2 = nMCCE2->get_electrometer_type();
        //- get type as string
        this->_electroCH2StrType = nMCCE2->get_electrometer_typeStr();
      }
    }
    catch(const electrometer::ElectrometerException& ke)
    {
      set_state(Tango::ALARM);
      set_status("Failed to find which Novelec type is connected on channel two : communication caught[...]");

      Tango::DevFailed df = electrometer_to_tango_exception(ke);
      FATAL_STREAM << "Failed to find which Novelec type is connected on channel two : ElectrometerException EXCEPTION -> " << df << std::endl;
    }
    catch(Tango::DevFailed& df)
    {
      set_state(Tango::ALARM);
      set_status("Failed to find which Novelec type is connected on channel two!");

      FATAL_STREAM << "Failed to find which Novelec type is connected on channel two : DevFailed EXCEPTION -> " << df << std::endl;
    }
    catch(...)
    {
      set_state(Tango::ALARM);
      set_status("Failed to find which Novelec type is connected on channel two : communication caught[...]");

      FATAL_STREAM << "Failed to find which Novelec type is connected on channel two : [...] EXCEPTION " << std::endl;
    }

    switch(type2)
    {
    case 0 :
      this->_electrometerCH2 = 0;
      break;
    case 1 :
    case 2 :
    case 3 :
      this->_electrometerCH2 = new N_PhotoVoltaique (proxyName, electrometer2Address,comProtocol);
      break;
    case 4 :
    case 5 :
      this->_electrometerCH2 = new N_PhotoConducteur(proxyName, electrometer2Address,comProtocol);
      break;
    default :
      this->_electrometerCH2 = 0;
      FATAL_STREAM << "\nERROR : Electrometers::create_electrometers_obj() -> this " << electrometer2Address << " Novelec type is not supported now!" << endl;
      return;
    }
    if ( mcce2 )
    {
      delete mcce2;
      mcce2 = 0;
    }
  }

  //- check
  if (!this->_electrometerCH1 && !this->_electrometerCH2)
  {
      set_state(Tango::FAULT);
      set_status("No electrometer connected on channels 1 and 2!");
    ERROR_STREAM << "No connected electrometer(s) found!" << std::endl;
    Tango::Except::throw_exception(
      "MEMORY_ERROR",
      "No connected electrometer(s) found!",
      "MCCE2Electrometers::create_electrometers_obj()");
  }

  //- configure MCCE-2 : clear error(s), explicit response, mode MEASURE OFF
  bool initCH1 = false;
  bool initCH2 = false;
  try
  {
    if ( this->_electrometerCH1 )
    {
      initCH1 = this->_electrometerCH1->init_protocol();
      if ( initCH1 )
        this->_electrometerCH1->mcce_init();

      //- read channel range
      this->_electrometer = this->_electrometerCH1;
      
      *attr_range1_read = this->get_range();
      
      INFO_STREAM << "this->_electrometer = this->_electrometerCH1!" << std::endl;
    }
    else
      WARN_STREAM << "WARN : !this->_electrometerCH1" << std::endl;

    if ( this->_electrometerCH2 )
    {
      initCH2 = this->_electrometerCH2->init_protocol();
      if ( initCH2 )
        this->_electrometerCH2->mcce_init();

      //- read channel range
      this->_electrometer = this->_electrometerCH2;

      *attr_range2_read = this->get_range();

      INFO_STREAM << "this->_electrometer = this->_electrometerCH2!" << std::endl;
    }

    //- start with channel one
    if ( this->_electrometerCH1 )
    {
      this->_electrometer = this->_electrometerCH1;
      this->_init_done = true;
    }
    //- if nothing connected on channel one, try with the second one
    else if ( this->_electrometerCH2 )
    {
      this->_electrometer = this->_electrometerCH2;
      this->_init_done = true;
    }
    else
      this->_init_done = false;
  }
  catch(const electrometer::ElectrometerException& ke)
  {
    Tango::DevFailed df = electrometer_to_tango_exception(ke);

    ERROR_STREAM << "Failed to set up the MCCE-2 module for a serial communication : ElectrometerException EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to set up the MCCE-2 module for a serial communication!",
                                      "MCCE2Electrometers::create_electrometers_obj()",
                                      Tango::ERR
                                     );
  }
  catch(Tango::DevFailed& df)
  {
    set_state(Tango::FAULT);
    set_status("Failed to set up the MCCE-2 module for a serial communication!");

    ERROR_STREAM << "Failed to set up the MCCE-2 module for a serial communication : DevFailed EXCEPTION -> " << df << std::endl;

    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_ERROR",
                                      "Failed to set up the MCCE-2 module for a serial communication!",
                                      "MCCE2Electrometers::create_electrometers_obj()",
                                      Tango::ERR
                                     );
  }
  catch(...)
  {
    set_state(Tango::FAULT);
    set_status("Failed to set up the MCCE-2 module for a serial communication caught[...]");

    ERROR_STREAM << "Failed to set up the MCCE-2 module for a serial communication : [...] EXCEPTION " << std::endl;

    Tango::Except::throw_exception(
      "UNKNOWN_ERROR",
      "Failed to set up the MCCE-2 module for a serial communication! [(...) exception]",
      "MCCE2Electrometers::create_electrometers_obj()",
      Tango::ERR
    );
  }
}


/*----- PROTECTED REGION END -----*/	//	MCCE2Electrometers::namespace_ending
} //	namespace
